/* Solution to the HPCU Bi-Weekly Challenge "Precipitate: Shared Memory
 * Implementation"
 * Skylar Thompson, University of Washington, and Aaron Weeden, Shodor
 * May 2014
 */

/* LIBS */
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <omp.h>

/* CONSTANTS */
/** States **/
const int NONE = -1;
const int PRECIPITATING = 0;
const int MOVING = 1;
const int BLOCKED = 2;
const int PRECIPITATE = 3;
const int COLLIDING = 4;

/** Directions **/
const int LEFT = 0;
const int RIGHT = 1;
const int UP = 2;
const int DOWN = 3;

/* INPUT */
int EnvWidth = 2000;
int EnvHeight = 2000;
int PartWidth = 1;
int PartHeight = 1;
int NParticles = 730;
int NTimeSteps = 2000;
int ChanceCollision = 50;

/* ARRAYS */
int *ParticleStates;
int *ParticleDirections;
int *XPositions;
int *YPositions;

/* FUNCTION DECLARATIONS */
void readData(int, char**);
void usage();
void randomizePositions(struct drand48_data *);
void pickDirections(struct drand48_data *);
bool tryMove();
void moveParticles();
void checkCollisions(struct drand48_data *);
bool checkPrecipitate();
void display();

/* MAIN */
int main(int argc, char **argv) {
    int i;
    struct drand48_data rand_buf;

    srand48_r((time(NULL)+getpid()),&rand_buf);
    
    readData(argc, argv);

    ParticleStates = (int*) malloc(NParticles * sizeof(int));
    ParticleDirections = (int*) malloc(NParticles * sizeof(int));
    XPositions = (int*) malloc(NParticles * sizeof(int));
    YPositions = (int*) malloc(NParticles * sizeof(int));

    randomizePositions(&rand_buf);
    for (i = 0; i < NTimeSteps; i++) {
        pickDirections(&rand_buf);
        moveParticles();
        checkCollisions(&rand_buf);
        display();
    }

    free(YPositions);
    free(XPositions);
    free(ParticleDirections);
    free(ParticleStates);

    exit(EXIT_SUCCESS);
}

/* FUNCTION DEFINITIONS */
void readData(int argc, char **argv) {
    int ch;

    while ((ch = getopt(argc, argv, "w:h:W:H:p:t:c:")) != -1) {
        switch (ch) {
            case 'w':
                EnvWidth = atoi(optarg);
                break;
            case 'h':
                EnvHeight = atoi(optarg);
                break;
            case 'W':
                PartWidth = atoi(optarg);
                break;
            case 'H':
                PartHeight = atoi(optarg);
                break;
            case 'p':
                NParticles = atoi(optarg);
                break;
            case 't':
                NTimeSteps = atoi(optarg);
                break;
            case 'c':
                ChanceCollision = atoi(optarg);
                break;
            case '?':
            default:
                usage(argv);
        }
    }
}

void usage(char **argv) {
    fprintf(stderr, "USAGE: %s [OPTIONS]\n", argv[0]);
    fprintf(stderr, "  OPTIONS:\n");
    fprintf(stderr, "    -w environment width\n");
    fprintf(stderr, "    -h environment height\n");
    fprintf(stderr, "    -W particle width\n");
    fprintf(stderr, "    -H particle height\n");
    fprintf(stderr, "    -p number of particles\n");
    fprintf(stderr, "    -t number of time steps\n");
    fprintf(stderr, "    -c chance of collision (integer between 0 (inclusive) and 100 (exclusive))\n");
}

void randomizePositions(struct drand48_data *rand_buf) {
    int i;
    double random_data;

    for (i = 0; i < NParticles; i++) {
        ParticleStates[i] = NONE;

        /* Generate a random double in [0,1) */
        drand48_r(rand_buf,&random_data);
        XPositions[i] = (EnvWidth - PartWidth) * random_data;

        drand48_r(rand_buf,&random_data);
        YPositions[i] = (EnvHeight - PartHeight) * random_data;
    }
}

void pickDirections(struct drand48_data *rand_buf) {
    int i;
    int dir;
    double random_data;

    for (i = 0; i < NParticles; i++) {
        if (ParticleStates[i] != PRECIPITATING) {
            drand48_r(rand_buf,&random_data);
            /* This will be a double casted back to an int */
            dir = 4 * random_data;
            if(tryMove(i,dir)) {
                ParticleDirections[i] = dir;
                ParticleStates[i] = MOVING;
            } else {
                ParticleStates[i] = BLOCKED;
            }
        }
    }
}

bool tryMove(int i, int dir) {
    int j;

    /* Can't move in a corner */
    if (dir == LEFT && XPositions[i] == 0) return false;
    else if (dir == RIGHT &&
            XPositions[i] == EnvWidth - PartWidth) return false;
    else if (dir == UP && YPositions[i] == 0) return false;
    else if (dir == DOWN &&
            YPositions[i] == EnvHeight - PartHeight) return false;

    /* Check to see if the non-corner particles have blocking neighbors */
    for (j = 0; j < NParticles; j++) {
        if (i != j) {
            if (dir == LEFT && XPositions[j] == XPositions[i] - 1 &&
                    YPositions[j] == YPositions[i]) return false;
            else if (dir == RIGHT && XPositions[j] == XPositions[i] + 1 &&
                    YPositions[j] == YPositions[i]) return false;
            else if (dir == UP && YPositions[j] == YPositions[i] - 1 &&
                    XPositions[j] == XPositions[i]) return false;
            else if (dir == DOWN && YPositions[j] == YPositions[i] + 1 &&
                    XPositions[j] == XPositions[i]) return false;
        }
    }

    return true;
}

void moveParticles() {
    int i;

#ifdef OPENMP
#pragma omp parallel for shared(ParticleStates, ParticleDirections, XPositions, YPositions)
#endif
    for (i = 0; i < NParticles; i++) {
        if (ParticleStates[i] == MOVING) {
            if (ParticleDirections[i] == LEFT)       XPositions[i]--;
            else if (ParticleDirections[i] == RIGHT) XPositions[i]++;
            else if (ParticleDirections[i] == UP)    YPositions[i]--;
            else if (ParticleDirections[i] == DOWN)  YPositions[i]++;
        } else {
            if (checkPrecipitate(i)) {
                ParticleStates[i] = PRECIPITATE;
            }
        }
    }
}

void checkCollisions(struct drand48_data *rand_buf) {
    int i;
    int j;
    double random_data;

#ifdef OPENMP
#pragma omp parallel for collapse(2) shared(ParticleStates, ParticleDirections, XPositions, YPositions)
#endif
    for (i = 0; i < NParticles; i++) {
        for (j = 0; j < NParticles; j++) {
            if (YPositions[i] == YPositions[j] &&
                    XPositions[i] == XPositions[j] - 1) {
                drand48_r(rand_buf,&random_data);
                if ((100 * random_data) < ChanceCollision) {
                    ParticleStates[i] = ParticleStates[j] = COLLIDING;
                    ParticleDirections[i] = ParticleDirections[j] = DOWN;
                }
            }
            if (ParticleStates[j] == PRECIPITATING &&
                    YPositions[j] == YPositions[i] - 1 &&
                    XPositions[i] == XPositions[j]) {
                ParticleStates[i] = PRECIPITATING;
                ParticleDirections[i] = DOWN;
            }
        }
    }
}

bool checkPrecipitate() {
    int i;
    int j;

    for (i = 0; i < NParticles; i++) {
        if (YPositions[i] == EnvHeight - PartHeight) return true;
        else
            for (j = 0; j < NParticles; j++)
                if (i != j)
                    if (ParticleStates[j] == PRECIPITATE &&
                            YPositions[j] == YPositions[i] + 1 &&
                            XPositions[j] == XPositions[i]) return true;
    }

    return false;
}

void display() {
    int i;
    int s;

    for (i = 0; i < NParticles; i++) {
        printf("particle %d: (%d,%d) ", i, XPositions[i], YPositions[i]);
        s = ParticleStates[i];
        if (s == PRECIPITATING) printf("precipitating\n");
        else if (s == MOVING) printf("moving\n");
        else if (s == BLOCKED) printf("blocked\n");
        else if (s == PRECIPITATE) printf("precipitate\n");
        else if (s == COLLIDING) printf("colliding\n");
    }
}
